//  Network.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Thu Apr 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "Network.h"


static unsigned long long nextID = 0;


void NetworkEntity::setNetwork(Network *newNetwork) {
    if (network != newNetwork) {
        vector<NetworkEntity*>::iterator iter, stopper;
        
        if (network) {
            network->removeEntity(this);
        }
        
        RELEASE(network);
        
        RETAIN(newNetwork);
        network = newNetwork;
        
        if (network) {
            network->addEntity(this);
        }
        
        // Note that the following may well result in this instance having setNetwork called back on it by one of the entities it calls against, as a result of loops in the network.  This is not a great issue though - when called back on this instance the 'newNetwork' paramater will already match this instances new network, so this function will return without further ado.  Thus, no infinite loops.
        
        for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
            (*iter)->setNetwork(network);
        }
    }
}

void NetworkEntity::unsetNetwork(Network *oldNetwork) {
    if (network && (network == oldNetwork)) {
        Network *temp = network;
        
        network = NULL;
        
        temp->removeEntity(this);
        
        RELEASE(temp);
    }
}

void NetworkEntity::submitToAllOutlets(const char *data, unsigned int length, bool &ownership, NetworkEntity *exception) {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
        if (*iter != exception) {
            (*iter)->submitData(data, length, this, ownership);
        }
    }
}

NetworkEntity::NetworkEntity(Network *home) : Object(), id(nextID++) {
    this->addType(NETWORK_ENTITY);
    
    if (home) {
        this->setNetwork(home);
    } else {
        network = NULL;
    }
};

NetworkEntity::~NetworkEntity() {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
        RELEASE(*iter);
    }
}

bool NetworkEntity::addOutlet(NetworkEntity *outlet) {
    vector<NetworkEntity*>::const_iterator iter, stopper;
    
    if (outlet && (outlet != this)) { // I supposed there could be a valid circumstance where connecting an outlet to ones-self could be useful, but conversely I know for certain it's very dangerous - real-time infinite loops may result.  So, the ability is disabled by default in this root class; subclasses could override this behaviour by forcefully adding the outlet as necessary.
        for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
            if (outlet == *iter) {
                return true;
            }
        }
        
        RETAIN(outlet);
        outlets.push_back(outlet);
        outlet->setNetwork(network);
        
        return true;
    } else {
        return false;
    }
}

void NetworkEntity::removeOutlet(NetworkEntity *outlet) {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
        if (outlet == *iter) {
            outlets.erase(iter);
            RELEASE(outlet);
            return;
        }
    }
}

const char* NetworkEntity::className(void) const {
    return "NetworkEntity";
}

void NetworkEntity::outputClassHierarchy(ostream &out) const {
    this->Object::outputClassHierarchy(out);
    out << " -> " << this->NetworkEntity::className();
}

void NetworkEntity::getDescription(ostream &out) const {
    this->Object::getDescription(out);
    out << "NetworkEntity ID: " << id << '\n';
    
    if (!outlets.empty()) {
        vector<NetworkEntity*>::const_iterator iter, stopper;
        
        out << "Outlets: ";
        
        for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
            out << (*iter)->getID();
            
            if ((iter + 1) != stopper) {
                out << ", ";
            }
        }
        
        out << '\n';
    }
}

void NetworkEntity::encode(Coder *coder) const {
    vector<NetworkEntity*>::const_iterator iter, stopper;

    this->Object::encode(coder);
    
    coder->encodeData((char*)&id, sizeof(unsigned long long)); // Just a cheap hack to encode a long long, which should be 64-bits (thus not an int), but may not be
    
    coder->encodeInt(outlets.size());
    
    for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
        coder->encodeObject(*iter);
    }
}

bool NetworkEntity::decode(Decoder *decoder) {
    int dataSize, outletCount;
    char *temp;
    
    if (this->Object::decode(decoder)) {
        if (!outlets.empty()) {
            fprintf(stderr, "Error: (%s) NetworkEntity[%llu] decoding failed - instance already has one or more outlets; previously initialised or decoded.\n", this->className(), id);
            return false;
        }
        
        temp = decoder->decodeData(dataSize);
        
        if ((dataSize != sizeof(unsigned long long)) || !temp) {
            fprintf(stderr, "Error: (%s) NetworkEntity[<%x>] decoding failed - invalid data size or no data for ID.\n", this->className(), (unsigned int)this);
            return false;
        }
        
        memcpy(&id, temp, sizeof(unsigned long long));
        
        free(temp);
        
        outletCount = decoder->decodeInt();
        
        if (outletCount < 0) {
            fprintf(stderr, "Error: (%s) NetworkEntity[%llu] decoding failed - invalid outlet count \"%d\".\n", this->className(), id, outletCount);
            return false;
        } else if (outletCount > 0) {
            NetworkEntity *current;
            
            outlets.reserve(outletCount);
            
            while (--outletCount >= 0) {
                current = dynamic_cast<NetworkEntity*>(decoder->decodeObject());
                
                if (!current) {
                    fprintf(stderr, "Error: (%s) NetworkEntity[%llu] decoding failed - unable to decode outlet object #%d.\n", this->className(), id, outlets.size());
                    return false;
                }
                
                outlets.push_back(current);
            }
        }
        
        return true;
    } else {
        return false;
    }
}


// Network implementation

void Network::setNetwork(Network *newNetwork) {
    if (newNetwork != this) {
        this->NetworkEntity::setNetwork(newNetwork);
    }
}

Network::Network(Network *home) : NetworkEntity(home), simulating(false) {
    this->addType(NETWORK);
}

Network::~Network() {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    for (iter = entities.begin(), stopper = entities.end(); iter != stopper; ++iter) {
        RELEASE(*iter);
    }
    
    for (iter = toRemove.begin(), stopper = toRemove.end(); iter != stopper; ++iter) {
        RELEASE(*iter);
    }
    
    for (iter = toAdd.begin(), stopper = toAdd.end(); iter != stopper; ++iter) {
        RELEASE(*iter);
    }
}

void Network::simulate(int time) {
    vector<NetworkEntity*>::iterator iter, stopper, outerIter, outerStopper;
    
    while (time--) {
        printf("--- (%s) Network[%llu] ticking entities. ---\n", this->className(), id);
        
        for (iter = entities.begin(), stopper = entities.end(); iter != stopper; ++iter) {
            (*iter)->tick();
        }
        
        for (outerIter = toRemove.begin(), outerStopper = toRemove.end(); outerIter != outerStopper; ++outerIter) {
            for (iter = entities.begin(), stopper = entities.end(); iter != stopper; ++iter) {
                if (*iter == *outerIter) {
                    (*iter)->unsetNetwork(this);
                    
                    RELEASE(*iter);
                    
                    iter = entities.erase(iter);
                }
            }
            
            RELEASE(*outerIter);
        }
            
        toRemove.clear();
        
        for (iter = toAdd.begin(), stopper = toAdd.end(); iter != stopper; ++iter) {
            entities.push_back(*iter);
            (*iter)->setNetwork(this);
        }

        toAdd.clear();
    }
}

bool Network::addEntity(NetworkEntity *entity) {
    vector<NetworkEntity*>::const_iterator iter, stopper;
    
    if (entity) {
        for (iter = entities.begin(), stopper = entities.end(); iter != stopper; ++iter) {
            if (entity == *iter) {
                return true;
            }
        }
        
        RETAIN(entity);
        
        if (simulating) {
            toAdd.push_back(entity);
        } else {
            entities.push_back(entity);
            entity->setNetwork(this);
        }
        
        return true;
    } else {
        return false;
    }
}

void Network::removeEntity(NetworkEntity *entity) {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    if (simulating) {
        toRemove.push_back(entity);
        RETAIN(entity);
    } else {
        for (iter = entities.begin(), stopper = entities.end(); iter != stopper; ++iter) {
            if (*iter == entity) {
                RELEASE(entity);
                entities.erase(iter);
                return;
            }
        }
    }
}

void Network::submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) {
    // Do nothing.  The generic Network instance has no idea how to handle data given to it - while it does have outlets as a result of it's NetworkEntity parentage, it does not make use of them in the normal way.  It is expected that Network subclasses will provide useful functionality in this regard, in their own particular ways.
}

void Network::tick(void) {
    this->simulate(1);
}

const char* Network::className(void) const {
    return "Network";
}

void Network::outputClassHierarchy(ostream &out) const {
    this->NetworkEntity::outputClassHierarchy(out);
    out << " -> " << this->Network::className();
}

void Network::getDescription(ostream &out) const {
    this->NetworkEntity::getDescription(out);
    
    if (!entities.empty()) {
        vector<NetworkEntity*>::const_iterator iter, stopper;
        
        out << "Entities: ";
        
        for (iter = entities.begin(), stopper = entities.end(); iter != stopper; ++iter) {
            out << (*iter)->getID();
            
            if ((iter + 1) != stopper) {
                out << ", ";
            }
        }
        
        out << '\n';
    }
}

void Network::encode(Coder *coder) const {
    vector<NetworkEntity*>::const_iterator iter, stopper;
    
    this->NetworkEntity::encode(coder);
    
    coder->encodeInt(entities.size());
    
    for (iter = entities.begin(), stopper = entities.end(); iter != stopper; ++iter) {
        coder->encodeObject(*iter);
    }    
}

bool Network::decode(Decoder *decoder) {
    int entityCount;
    
    if (this->NetworkEntity::decode(decoder)) {
        if (!entities.empty()) {
            fprintf(stderr, "Error: (%s) Network[%llu] decoding failed - instance already has one or more entities; previously initialised or decoded.\n", this->className(), id);
            return false;
        }
        
        entityCount = decoder->decodeInt();
        
        if (entityCount < 0) {
            fprintf(stderr, "Error: (%s) Network[%llu] decoding failed - invalid entity count \"%d\".\n", this->className(), id, entityCount);
            return false;
        } else if (entityCount > 0) {
            NetworkEntity *current;
            
            entities.reserve(entityCount);
            
            while (--entityCount >= 0) {
                current = dynamic_cast<NetworkEntity*>(decoder->decodeObject());
                
                if (!current) {
                    fprintf(stderr, "Error: (%s) Network[%llu] decoding failed - unable to decode entity #%d.\n", this->className(), id, entities.size());
                    return false;
                }
                
                entities.push_back(current);
            }
        }
        
        return true;    
    } else {
        return false;
    }
}
