//  Network.h
//  Network Simulator
//
//  Created by Wade Tregaskis on Thu Apr 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NETWORK_H
#define NETWORK_H

#include "Object.h"

#include <vector>
using namespace std;


#define NETWORK_ENTITY 1
#define NETWORK 2


class Network;


/*! @class NetworkEntity
    @abstract An abstract representation of some sort of network component, e.g. a transmission link, a router, a server, etc.
    @discussion The NetworkEntity class is designed to represent the lowest common denominator of all network devices.  It is the foundation block upon which all simulation components are built, and from which all simulations are constructed.  It does nothing useful by itself.
    */

class NetworkEntity : public Object {
protected:
    Network *network; // The network we are a part of
    vector<NetworkEntity*> outlets; // Outlets we connect to
    unsigned long long id; // Our GUID
    
/*! @function submitToAllOutlets
    @abstract Submits the given data to all the receiver's outlets.
    @discussion This submits the given data to each one of the receiver's outlets in sequence.  The source given to those outlets in the submitData member function is automatically 'this'.  The ownership flag works exactly the same as when calling submitData normally - if 'true' on input any one of the outlets may take control of it, in which case it will be 'false' on output.  If 'false' on input, it will not change and will be returned as 'false'.  Note that each outlet probably would prefer to take ownership of the data, for efficiency, but this cannot happen.  Once one outlet takes control of the data, all others are forced to copy it.  There's no real way of avoiding this [in the current design].
    @param data The data to submit to each outlet.  See the discussion for the submitData member function for more information.
    @param length The length of the data being submitted.  See the discussion for the submitData member function for more information.
    @param ownership The ownership of the data going in and out of this member function.  See the discussion for the submitData member function for more information.
    @param exception A NetworkEntity to not send the data to, if it happens to appear in the outlets list.  This is obviously useful on bi-directional links where you don't want to echo the data back to the sender. */
    
    virtual void submitToAllOutlets(const char *data, unsigned int length, bool &ownership, NetworkEntity *exception = NULL);
    
public:
/*! @function NetworkEntity
    @abstract Designated initialiser for the NetworkEntity class.
    @discussion Initialises the receiver as appropriate.  If the optional parameter is provided, the receiver is added to the given network automatically.  If none is provided, the receiver is initialised as belong to no particular network; one may later be assigned using the setNetwork member function.
    @param home The Network to automatically add the receiver to, if any. */
        
    NetworkEntity(Network *home = NULL);
    ~NetworkEntity(); // Standard destructor, a block box from anyone's point of view; no documentation necessary
    
/*! @function setNetwork
    @abstract Sets the network of the receiver.
    @discussion In order for a NetworkEntity to be simulated, it usually must belong to a Network.  It is normal practice when creating NetworkEntity's to assign them to a network in the constructor.  Very important: This member function also changes the network of all the receiver's current outlets to be that of the new network - i.e. two NetworkEntity's connected via an outlet may not be on separate networks.  This raises some troublesome design issues with regards to allowing multiple layers of Networks (i.e. Networks containing other Networks, ad infinum).  There are no guarantee's at time of writing that the system works in this ideal fashion; i.e. it's untested and there are potential pitfalls.
    @param newNetwork The new network to place the receiver in.  The receiver is automatically removed from the old network, if any.  This may be NULL. */
    
    virtual void setNetwork(Network *newNetwork);
    
/*! @function unsetNetwork
    @abstract Unsets the given network as the network of the receiver.
    @discussion This differs from setNetwork(NULL) in an important way - it only sets the network to NULL if it was already set to the given network.  The idea is that when removing from a network you'll want to unset the network of each entity in the network.  But some of those may have already moved to other networks.  So, you only want to clear their network setting if it's still on the old network.
    @param oldNetwork The old network to remove the receiver from, if it is on that network. */
    
    virtual void unsetNetwork(Network *oldNetwork);

/*! @function canAcceptData
    @abstract Allows the receiver to indicate whether it can accept data of the given length from a particular source.
    @discussion Since submitData has no idea what the receiver does with the given data, this member function allows you to query the receiver to see if it actually wants it or not.
    @param length The length of the hypothetical data.
    @param source The source from which the data will be sent, which may be NULL.
    @result Returns true if the receiver is [at this exact moment in time] willing to receive data of the given length from the given source [if any], false otherwise. */
        
    virtual bool canAcceptData(int length, NetworkEntity *source) { return false; };
    
/*! @function submitData
    @abstract Submits data to the receiving NetworkEntity.
    @discussion This function hands some data to the receiver, which does whatever it likes with it.
    @param data The data to submit.
    @param length The length of the data.
    @param source The NetworkEntity submitting the data to the receiver.
    @param ownership As an input tells the receiver whether or not it may take ownership of the data.  That is, whether or not it can use the data directly (rather than copying it) and is thus responsible for freeing it, rather than the caller.  If the receiver wishes to take ownership, it need only toggle the value to false, which as an output informs the caller that it no longer has ownership of the data.  If the value of this going into the member function is false, it must be false on return.
    @result Returns the transmission time of the given data. */
    
    virtual void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) = 0;
    
/*! @function tick
    @abstract Allows the receiver to carry out any operations necessary for the current unit time step.
    @discussion NetworkEntity's are stepped over time by having tick called on all their objects.  This is usually managed by their containing Network instance.  The receiver of the tick should call submitData on any of it's outlets, as necessary, and perform whatever else it needs to do for the current time step.
    @param out The output stream to which any information should be written. */
        
    virtual void tick(void) = 0;
    
/*! @function addOutlet
    @abstract Adds a NetworkEntity as an outlet of the receiver.
    @discussion NetworkEntity's communicate by passing data to their outlets.  Outlets are one-way streams; two-way connections may be established by adding both objects as outlets of each other.
        
                Note that the given outlet will have it's network changed, if necessary, to be the same as that of the receiver.  Consequently, be careful not to connect NetworkEntity's between layers - the result of such an event is reliably catastrophic.
    @param outlet The outlet to add.
        
                Also note that subclasses of NetworkEntity may place restrictions on the number and types of outlets that can be added.  For example, the NetworkLink class allows only a single outlet; attempts to add more will fail.  Other classes may only allow outlets to a certain type of NetworkEntity.  The morale is, never ever expect this operation to just be successful - always check the return value and act appropriately.
    @result Returns true if the outlet was added, false otherwise.  If the receiver already has an outlet to the given object, true is returned anyway (well, it was added at *some* point). */
    
    virtual bool addOutlet(NetworkEntity *outlet);

/*! @function removeOutlet
    @abstract Removes a NetworkEntity from the receiver's list of outlets.
    @discussion The given outlet is removed from the receiver's list of outlets.  Nothing more is changed - the given outlet's network will remain the same, even if it had been previously set/changed as a result of being made the outlet of the receiver.  NetworkEntity's can only be completely removed from a Network using the appropriate member function of the Network class.
    @param outlet The outlet to remove. */
    
    virtual void removeOutlet(NetworkEntity *outlet);
    
/*! @function getID
    @abstract Returns the globally unique id of the receiver.
    @discussion When a NetworkEntity is created it assigns itself a globally unique ID [within the current process].  This allows you to identify NetworkEntity's at any point during a program's execution.
        
                Note: there is no real error checking to ensure the ID's are globally unique, but the system is unlikely to fail in practice - unless you happen to allocate over eighteen billion billion (yes, billion billion) NetworkEntity's in one run.  Go on, do your worst. :P
        
                P.S. If you're wondering why an arbitrary ID has to be used, and objects can't just be identified by their memory address as might happen in other circumstances, the reason is that over time NetworkEntity's may be destroyed and more created, possibly in the same location.  Thus, a memory address may point to multiple objects (and other random data!) over time.  So, while memory addresses are used for identification in certain places (i.e. the standard Object coding/decoding system), in any case where there is a temporal element involved, these ID's are used instead.
    @result Returns the globally unique ID of the receiver. */
        
    unsigned long long getID(void) const { return id; }
    
    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};


/*! @class Network
    @abstract Represents and simulates a Network, containing any number of NetworkEntity's.
    @discussion In order to efficiently and simply simulate a network of any useful complexity, it is convenient to arrange all the NetworkEntity's under the control of a single object, which can then handle some of the details of the simulation - things like stepping each object during each simulation cycle, for example.  The Network class performs this functionality in a basic way.  NetworkEntity's added to it are ticked once per simulation cycle, in the order they were initially added.
                Interestingly, as a subclass of NetworkEntity itself, the Network class can be used as a NetworkEntity in other networks.  At least, this is the design intention.  At present the base Network class does not offer the functionality to do this, but it is entirely possible through the use of the appropriate purpose-built subclasses (i.e. the generic Network has no concept of it's interfaces, so it doesn't know what to do with data submitted to it, nor is it capable of receiving data from entities within itself).
    */

class Network : public NetworkEntity {
protected:
    vector<NetworkEntity*> entities, toRemove, toAdd; // It's dangerous to add/remove entities to the network during simulation, as there is the potential for crashing and infinite loops.  Consequently all such operations are delayed until the end of the current simulation cycle [if any], thus the need for the toRemove and toAdd vectors, in addition to the self-explanatory entities vector.
    bool simulating; // True if the receiver is currently performing a simulation, and is within a particular cycle of that simulation.  Used in to control the timing of entity addition/removal, as discussed above.
    
    void setNetwork(Network *newNetwork); // override the default NetworkEntity version
    
public:
/*! @function Network
    @abstract Designated initialiser for the Network class.
    @discussion Initialises the Network as appropriate.  As a NetworkEntity subclass the receiver can be given another Network for which it itself can be a part of, although as mentioned in the class discussion there is no real support for this in the Network base class itself; purpose-built and suited subclasses are expected to provide this.
    @param home The Network to automatically add the receiver to, if any. */
        
    Network(Network *home = NULL);
    ~Network(); // Standard destructor, a block box from anyone's point of view; no documentation necessary
    
/*! @function simulate
    @abstract Performs the given number of consecutive simulation cycles of the receiver.
    @discussion Fairly self-explanatory.  Each cycle involves the 'ticking' of each entity in the network once and only once, which allows those entities to perform their function.  Entities are ticked in the order they were added to the receiver, which is a significant thing to note - entities interact with each other during their tick, which means the behaviour of ticking one entity followed by another may not be the same as ticking them in reverse order.  Be mindful of this.
    @param time The number of simulation cycles to perform.  Must be greater than or equal to 1. */
    
    virtual void simulate(int time);
    
/*! @function addEntity
    @abstract Adds a given NetworkEntity to the receiver.
    @discussion The given NetworkEntity is added to the receiver, if it is not already in the receiver.  A NetworkEntity can only belong to one network at once, and as such this member function automatically sets the owner network of the given NetworkEntity to itself.  This may cause any other NetworkEntity's linked to the given NetworkEntity to also be moved into the receiver - see the relevant documentation for the NetworkEntity setNetwork member function for more details.
    @param entity The entity to add to the receiver.
    @result Returns true if the given NetworkEntity was successfully added to the receiver [or was already a member of], false otherwise. */
    
    virtual bool addEntity(NetworkEntity *entity);
    
/*! @function removeEntity
    @abstract Removes a given NetworkEntity from the receiver.
    @discussion If the given NetworkEntity exists within the receiver, it is removed and it's owning network set to NULL.  This may result in multiple objects being removed from the receiver, due to the propogation of network addition/removal - see the discussion for the addEntity member function for more details.
    @param entity The entity to remove from the receiver. */
    
    virtual void removeEntity(NetworkEntity *entity);
    
    // NetworkEntity
    
    void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership);
    void tick(void);
    
    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};

#endif
