//  NetworkFileSink.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Fri Apr 02 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "NetworkFileSink.h"


char* debugBuffer;
unsigned int debugBufferLength;


NetworkFileSink::NetworkFileSink(Network *home, const char *file) : NetworkEntity(home) {
    this->addType(NETWORK_FILE_SINK);
    
    assert(file);
    
    int length = strlen(file) + 1;
    
    fileName = (char*)malloc(length);
    memcpy(fileName, file, length);
    
    output.clear();
    output.open(file, ios::out);
}

NetworkFileSink::~NetworkFileSink() {
    free(fileName);
    output.close();
}

bool NetworkFileSink::canAcceptData(int length, NetworkEntity *source) {
    return output.good();
}

void NetworkFileSink::submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) {
    assert(data != NULL);
    
    if (output.good()) {
#ifndef NDEBUG
        // This debug code is the ultimate line of defense against errors - it picks up any difference between the sourced data and sinking data in any network containing only a single source/sink combo.
        if (debugBuffer == NULL) {
            fflush(stdout);
            fprintf(stderr, "\nPacket sourced data is non-existant.\n\n");
            fflush(stderr);
            exit(-5);
        } else if (debugBufferLength != length) {
            fflush(stdout);
            fprintf(stderr, "\nPacket sourced length doesn't match sinking length.\n\n");
            fflush(stderr);
            exit(-6);
        } else if (strncmp(debugBuffer, data, length) != 0) {
            fflush(stdout);
            fprintf(stderr, "\nPacket sourced data doesn't match sinking data.\n\n");
            fflush(stderr);
            exit(-7);
        }
#endif
        
        printf("(%s) NetworkFileSink[%llu] wrote %d bytes to file.\n", this->className(), id, length);
        output.write(data, length);
        output.flush(); // We have to flush after each write, as inefficient as it may be, because we may not get properly released due to the complicated circular dependencies in the network construct, thus, the destructor may never be called -> file may never actually be closed -> no guarantee of penultimate flush.
    } else {
        fprintf(stderr, "Error: (%s) NetworkFileSink[%llu] unable to write to file because it's not good.\n", this->className(), id);
    }
}

void NetworkFileSink::tick(void) {
    // Do nothing
}

bool NetworkFileSink::addOutlet(NetworkEntity *outlet) {
    return false;
}

// Object

const char* NetworkFileSink::className(void) const {
    return "NetworkFileSink";
}

void NetworkFileSink::outputClassHierarchy(ostream &out) const {
    this->NetworkEntity::outputClassHierarchy(out);
    out << " -> " << this->NetworkFileSink::className();
}

void NetworkFileSink::getDescription(ostream &out) const {
    this->NetworkEntity::getDescription(out);
    out << "Output file: " << (fileName ? fileName : "(none)") << '\n';
}

void NetworkFileSink::encode(Coder *coder) const {
    this->NetworkEntity::encode(coder);
    
    coder->encodeString(fileName);
    coder->encodeInt(const_cast<NetworkFileSink*>(this)->output.tellp());
}

bool NetworkFileSink::decode(Decoder *decoder) {
    int fileLength;
    
    if (this->NetworkEntity::decode(decoder)) {
        fileName = decoder->decodeString();
        
        assert(fileName);
        
        output.clear();
        output.open(fileName, ios::ate | ios::out);
        
        fileLength = decoder->decodeInt();
        
        if (!output.good() || (fileLength != output.tellp())) {
            fprintf(stderr, "Error: (%s) NetworkFileSink[%llu] decoding failed - the output file it originally pointed to either no longer exists or has changed - \"%s\".\n", this->className(), id, fileName);
            return false;
        }
        
        return true;
    } else {
        return false;
    }
}
