//  NetworkFileSink.h
//  Network Simulator
//
//  Created by Wade Tregaskis on Fri Apr 02 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NETWORK_FILE_SINK_H
#define NETWORK_FILE_SINK_H

#include "Network.h"


#define NETWORK_FILE_SINK 6


/*! @class NetworkFileSink
    @abstract Represents a point in the network at which all data received is written to a particular file.
    @discussion The NetworkFileSink is a very simple creature - it takes all data submitted to it from other NetworkEntity's and writes it to the file it represents.  That's it.  It is of course complemented by the NetworkFileSource class, which performs the opposite operation.
                Note that the NetworkFileSink writes data in the order it receives it.  If multiple sources are capable of sending data to the sink, it is the responsability of other network components to order that data as necessary.
    */

class NetworkFileSink : public NetworkEntity {
private:
    char *fileName; // The file name we write to... kept here only for stat() purposes, which would be redundant if we used standard C file operations instead of an ofstream... but I wrote it that way to begin with and am too lazy to change it now; it works well enough.
    ofstream output; // The output stream to which we write.

public:
/*! @function NetworkFileSink
    @abstract Designated initialiser for the NetworkFileSink class.
    @discussion Initialises the receiver for operation within the given network, writing to the file at the given path.
    @param home The network to automatically add the receiver to.  Although this is not an optional parameter here, you can still pass NULL explicitly.
    @param file The path of the file to write data to.  If a file already exists at this location it is replaced.  Note that if the file cannot be opened or created for whatever reason, the sink will be initialised but inoperable - all data submitted to it will log an error message and be ignored. */
        
    NetworkFileSink(Network *home, const char *file);
    ~NetworkFileSink(); // Standard block box destructor... nothing to see here.
  
    // NetworkEntity
    
    bool canAcceptData(int length, NetworkEntity *source);
    void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership);
    void tick(void);
    bool addOutlet(NetworkEntity *outlet);

    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};

#endif
