//  NetworkFileSource.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Fri Apr 02 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "NetworkFileSource.h"


extern char* debugBuffer;
extern unsigned int debugBufferLength;


NetworkFileSource::NetworkFileSource(Network *home, const char *file, int dataSize, int timeBetween) : NetworkEntity(home), frameDataSize(dataSize), timeBetweenFrames(timeBetween), buffer(NULL) {
    this->addType(NETWORK_FILE_SOURCE);

    assert(file);
    assert(frameDataSize >= 0);
    assert(timeBetweenFrames >= 0);
    
    int length = strlen(file) + 1;
        
    fileName = (char*)malloc(length);
    memcpy(fileName, file, length);
    
    input.clear();
    input.open(file);
}

NetworkFileSource::~NetworkFileSource() {
    if (fileName) {
        free(fileName);
    }
    
    if (buffer) {
        free(buffer);
    }
    
    input.close();
}

bool NetworkFileSource::finished(void) {
    return !input.good();
}

bool NetworkFileSource::canAcceptData(int length, NetworkEntity *source) {
    printf("Warning: NetworkFileSource[%llu] was asked if canAcceptData, when it really shouldn't be (i.e. it implies it's the outlet to something, which is Bad)", id);
    
    return false;
}

void NetworkFileSource::submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) {
#ifndef NDEBUG
    printf("Warning: NetworkFileSource[%llu] given data via submitData (ignoring) - make sure it's not the outlet of something", id);
#endif
    
    return;
}

void NetworkFileSource::tick(void) {
    if (input.good()) {
        if (0 == timeTillNextFrame--) {
            vector<NetworkEntity*>::iterator iter, stopper;
            bool ownership = true;
            int size;
            
            if (frameDataSize == 0) {
                struct stat sb;
                
                assert(0 == stat(fileName, &sb));
                
                size = sb.st_size;                
            } else {
                size = frameDataSize;
            }
            
            for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
                if (!(*iter)->canAcceptData(frameDataSize, this)) {
                    //printf("(%s) NetworkFileSource[%llu] - Unable to source data because outlet %llu is not ready.\n", this->className(), id, (*iter)->getID());
                    timeTillNextFrame = 0;
                    return;
                }
            }
            
            if (!buffer) {
                buffer = (char*)malloc(size);
            }
            
            assert(buffer); // I wouldn't normally bother checking, but we might be trying to read a truly massive file (if frameDataSize == 0 or otherwise very big, so it's worth ensuring the malloc worked

            input.read(buffer, size);
            size = input.gcount();
            
#ifndef NDEBUG
            if (debugBuffer) {
                free(debugBuffer);
            }
            
            debugBuffer = (char*)malloc(size);
            memcpy(debugBuffer, buffer, size);
            debugBufferLength = size;
#endif
            
            if (size > 0) {
                printf("(%s) NetworkFileSource[%llu] sourcing %d bytes.\n", this->className(), id, size);
                
                this->submitToAllOutlets(buffer, size, ownership);
            }
            
            if (!ownership) { // If it *was* taken over by someone else, we'll have to clear our reference to it and create a new one when needed (next time round)
                buffer = NULL;
            }
            
            timeTillNextFrame = timeBetweenFrames;
        }
    }
}

const char* NetworkFileSource::className(void) const {
    return "NetworkFileSource";
}

void NetworkFileSource::outputClassHierarchy(ostream &out) const {
    this->NetworkEntity::outputClassHierarchy(out);
    out << " -> " << this->NetworkFileSource::className();
}

void NetworkFileSource::getDescription(ostream &out) const {
    this->NetworkEntity::getDescription(out);
    out << "Source file: " << (fileName ? fileName : "(none)") << "\nFrame data size: " << frameDataSize << "\nTime between frames: " << timeBetweenFrames << "\nTime till next frame: " << timeTillNextFrame << '\n';
}

void NetworkFileSource::encode(Coder *coder) const {
    this->NetworkEntity::encode(coder);
    
    coder->encodeInt(frameDataSize);
    coder->encodeInt(timeBetweenFrames);
    coder->encodeInt(timeTillNextFrame);
    
    coder->encodeString(fileName);
    coder->encodeInt(const_cast<NetworkFileSource*>(this)->input.tellg()); // Dumbass stdlib not providing proper const member functions
}

bool NetworkFileSource::decode(Decoder *decoder) {
    if (this->NetworkEntity::decode(decoder)) {    
        frameDataSize = decoder->decodeInt();
        timeBetweenFrames = decoder->decodeInt();
        timeTillNextFrame = decoder->decodeInt();
        
        fileName = decoder->decodeString();
        
        if (!fileName) {
            fprintf(stderr, "Error: (%s) NetworkFileSource[%llu] decoding failed - source filename missing.\n", this->className(), id);
            return false;
        }
        
        input.clear();
        input.open(fileName);
        
        input.seekg(decoder->decodeInt());
        
        return true;
    } else {
        return false;
    }
}
