//  NetworkFileSource.h
//  Network Simulator
//
//  Created by Wade Tregaskis on Fri Apr 02 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NETWORK_FILE_SOURCE_H
#define NETWORK_FILE_SOURCE_H

#include "Network.h"

#include <sys/types.h>
#include <sys/stat.h>

#include <fstream>
using namespace std;


#define NETWORK_FILE_SOURCE 4


/*! @class NetworkFileSource
    @abstract Represents a source of data into the network, by reading data in a linear fashion from a particular file.
    @discussion The complementary opposite of the NetworkFileSink, this class reads data from a particular file and transmits it to it's outlet(s).  It does so only when <b>all</b> it's outlets are ready, and only in blocks equal to or less than a given size (less than in the case that the end of file is reached without a full block being read; otherwise the blocks are guaranteed to be full).
    */

class NetworkFileSource : public NetworkEntity {
private:
    int frameDataSize, timeBetweenFrames, timeTillNextFrame; // frameDataSize is how many bytes to read at a time, timeBetweenFrames is the minimum delay between sourcing (which does not take into effect any ticks in which the receiver(s) were not ready, and timeTillNextFrame is used in conjunction with the 'constant' timeBetweenFrames member variable to count down each tick, until data can be sourced.
    char *fileName; // The file name to source data from.
    char *buffer; // Buffer to read data into
    ifstream input; // The input stream to source data from.
    
public:
/*! @function NetworkFileSource
    @abstract Designated initialiser for the NetworkFileSource class.
    @discussion Initialises the receiver as a file source on the given network, for the given file.
    @param home The network to initialise the receiver in (may be NULL).
    @param file The file to source the data from.  This must exists and be readable; no error checking is performed.  This may not be NULL.
    @param dataSize How much data to send from the file at a time (per frame).  If it is 0, the entire contents of the frame is dumped to the outlets at once.
    @param timeBetween The time between frames.  If 0, a frame is sent on every tick.  If 1, a frame is sent every other tick, etc. */
        
    NetworkFileSource(Network *home, const char *file, int dataSize = 0, int timeBetween = 0);
    ~NetworkFileSource(); // Standard block box destructor... nothing to see here.
    
/*! @function finished
    @abstract Returns whether or not the receiver has finished sourcing the file.
    @discussion The sourcing is considered complete if an error occurs, or end-of-file is reached, or in any other way it is not possible - and never will be possible - to source any more data.
    @result True if the receiver has finished sourcing (and will send no further data to it's outlets), false otherwise. */
    
    bool finished(void);
    
    // NetworkEntity
    
    bool canAcceptData(int length, NetworkEntity *source);
    void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership);
    void tick(void);
    
    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};

#endif

