//  NetworkLink.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Thu Apr 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "NetworkLink.h"


NetworkLink::NetworkLink(Network *home, int transmission) : NetworkEntity(home), transmissionTime(transmission) {
    this->addType(NETWORK_LINK);
    
    assert(transmission >= 0);
}

NetworkLink::~NetworkLink() {
    vector< pair< pair<char *, unsigned int>, int> >::iterator iter, stopper;

    for (iter = dataInTransit.begin(), stopper = dataInTransit.end(); iter != stopper; ++iter) {
        free(iter->first.first);
    }
}

bool NetworkLink::addOutlet(NetworkEntity *outlet) {
    if (outlets.empty()) {
        return this->NetworkEntity::addOutlet(outlet);
    } else {
        return false;
    }
}

bool NetworkLink::canAcceptData(int length, NetworkEntity *source) {
    return true;
}

void NetworkLink::submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) {
    char *myData;

    if (ownership) {
        ownership = false;
        myData = const_cast<char*>(data); // Muahaha, we are so evil
    } else {
        myData = (char*)malloc(length);
        memcpy(myData, data, length);
    }

    if (transmissionTime > 0) {
        dataInTransit.push_back(pair<pair<char*, unsigned int>, int>(pair<char*, unsigned int>(myData, length), transmissionTime)); // Eek... if you need any further proof that C++ & it's oft-praised templates are bad...
    } else {
        bool myOwnership = true;
        
        this->submitToAllOutlets(myData, length, myOwnership);
        
        if (myOwnership) {
            free(myData);
        }
    }
}

void NetworkLink::tick(void) {
    vector< pair< pair<char *, unsigned int>, int> >::iterator iter, stopper;
    bool ownership;
    
    for (iter = dataInTransit.begin(), stopper = dataInTransit.end(); iter != stopper; ++iter) {
        --(iter->second);
        
        if (iter->second <= 0) {
            ownership = true;
            
            this->submitToAllOutlets(iter->first.first, iter->first.second, ownership, NULL);
            
            if (ownership) {
                free(iter->first.first);
            }
            
            iter = dataInTransit.erase(iter);
        }
    }
}

const char* NetworkLink::className(void) const {
    return "NetworkLink";
}

void NetworkLink::outputClassHierarchy(ostream &out) const {
    this->NetworkEntity::outputClassHierarchy(out);
    out << " -> " << this->NetworkLink::className();
}

void NetworkLink::getDescription(ostream &out) const {
    this->NetworkEntity::getDescription(out);
    out << "Transmission time: " << transmissionTime << '\n';
    // FLAG - we currently don't output the data in ourselves, which we should... but how to uniquely identify it?
}

void NetworkLink::encode(Coder *coder) const {
    vector< pair< pair<char *, unsigned int>, int> >::const_iterator iter, stopper;
    pair< pair<char *, unsigned int>, int> current;
    
    this->NetworkEntity::encode(coder);
    
    coder->encodeInt(transmissionTime);
    coder->encodeInt(dataInTransit.size());
    
    for (iter = dataInTransit.begin(), stopper = dataInTransit.end(); iter != stopper; ++iter) {
        current = *iter;
        
        coder->encodeData(current.first.first, current.first.second);
        coder->encodeInt(current.second);
    }
}

bool NetworkLink::decode(Decoder *decoder) {
    if (this->NetworkEntity::decode(decoder)) {
        int size;
        char *data;
        int dataLength, timeLeft;
        
        transmissionTime = decoder->decodeInt();
        
        size = decoder->decodeInt();
        
        while (size--) {
            data = decoder->decodeData(dataLength);
            
            if (dataLength < 0) {
                fprintf(stderr, "Error: (%s) NetworkLink[%llu] decoding failed - invalid data length \"%d\".\n", this->className(), id, dataLength);
                return false;
            }
            
            timeLeft = decoder->decodeInt();
            
            dataInTransit.push_back(pair<pair<char*, unsigned int>, int>(pair<char*, unsigned int>(data, dataLength), timeLeft));
        }
        
        return true;
    } else {
        return false;
    }
}
