//  NetworkLink.h
//  Network Simulator
//
//  Created by Wade Tregaskis on Thu Apr 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NETWORK_LINK_H
#define NETWORK_LINK_H

#include "Network.h"


#define NETWORK_LINK 3


#define RANDOM(min, max) ((min) + (((max) - (min)) * ((float)random() / RAND_MAX)))


/*! @class NetworkLink
    @abstract Represents a generic network link with an optional transmission delay.
    @discussion The NetworkLink object receives data from one or more sources and forwards it, after some configured delay, to the outlet.  Note that a NetworkLink may have one and only one outlet; attempts to add more than one will fail, returning false as appropriate.  Multiple sets of data may be in flight at any one time - they will all come out in the correct order and timing.  Note that the output is written to the outlet regardless of whether or not the outlet is willing to accept it, so data may be lost if the outlet does not handle it.  Use an appropriate buffer if necessary, or better yet simply design your network such that the outlets from the link are always ready.
    */

class NetworkLink : public NetworkEntity {
protected:
    int transmissionTime; // The time in ticks between data entry and it's subsequent exit.
    
    vector< pair< pair<char *, unsigned int>, int> > dataInTransit; // The inner pair is the data and it's length, embedded in another pair where the other parameter is the amount of time left before the data gets sent to the outlet
        
public:
/*! @function NetworkLink
    @abstract Designated initialiser for the NetworkLink class.
    @discussion Initialises the receiver in the given network with the given properties.
    @param transmission The transmission time for data across the link.  A transmission time of 0 has the link pass all data as soon as it is received (i.e. in the same tick).  A transmission time of 1 has it pass data out on the tick after it receives it, etc etc.  Must be greater than or equal to 0. */
        
    NetworkLink(Network *home = NULL, int transmission = 0);
    ~NetworkLink(); // Standard block box destructor... nothing to see here.
    
    // NetworkEntity
        
    bool addOutlet(NetworkEntity *outlet);
    bool canAcceptData(int length, NetworkEntity *source);
    void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership);
    void tick(void);
    
    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};

#endif
