//  NetworkLinkCorruptor.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Thu Apr 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "NetworkLinkCorruptor.h"


// The true probabilistic method for corrupting the data.  This is probably much slower than the statistically-valid but predictable version.
void NetworkLinkCorruptor::toggleBits(char *data, unsigned int length) {
    unsigned int theByte = 0, bitOfByte = 0;
    char mask;
    
    while (theByte < length) {
        while (bitOfByte < 8) {
            if (RANDOM(0.0, 1.0) < bitToggleProbability) {
                mask = 1 << bitOfByte;
                
                data[theByte] ^= mask;
                
                printf("(%s) NetworkLinkCorruptor[%llu] toggling bit %d of byte %d of data.\n", this->className(), id, bitOfByte, theByte);
            }
            
            ++bitOfByte;
        }
        
        ++theByte;
        bitOfByte = 0;
    }
}

// This being that predictable version - it tends to toggle the bits at regular time intervals, since it's just counting the bits as they pass by and toggling some random bit around every Nth bit.  It will never, if the packet is small enough, toggle more than one bit per packet.
/*void NetworkLinkCorruptor::toggleBits(char *data, unsigned int length) {
    static float numberOfBitsToToggle = 0;
    int exactBit, theByte, bitOfByte;
    char mask;

    numberOfBitsToToggle += (length * 8) * bitToggleProbability;
    
    while (numberOfBitsToToggle >= 1.0) {
        exactBit = (int)round(RANDOM(0.0, (length * 8) - 1));
        
        theByte = exactBit / 8;
        bitOfByte = exactBit % 8;
        
        printf("(%s) NetworkLinkCorruptor[%llu] toggling bit %d of byte %d of data.\n", this->className(), id, bitOfByte, theByte);
        
        //mask = 1 << (bitOfByte - 1); // Surely this is wrong, since bitOfByte is in the range [0, 7], so sometimes this would try to shift the 1 by -1... which might result in 0.
        mask = 1 << bitOfByte;

        data[theByte] ^= mask;
        
        numberOfBitsToToggle -= 1.0;
    }
}*/

NetworkLinkCorruptor::NetworkLinkCorruptor(Network *home, int transmission, float bitToggleProb) : NetworkLink(home, transmission), bitToggleProbability(bitToggleProb) {
    this->addType(NETWORK_LINK_CORRUPTOR);
    
    assert(bitToggleProb >= 0.0);
}

NetworkLinkCorruptor::~NetworkLinkCorruptor() {
    // Do nothing
}

void NetworkLinkCorruptor::submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) {
    char *myData;
    bool myOwnership = true;

    if (ownership) {
        ownership = false;
        myData = const_cast<char*>(data); // Muahaha, we are so evil
    } else {
        myData = (char*)malloc(length);
        memcpy(myData, data, length);
    }
    
    this->toggleBits(myData, length);

    this->NetworkLink::submitData(myData, length, source, myOwnership);
    
    if (myOwnership) {
        free(myData);
    }
}

const char* NetworkLinkCorruptor::className(void) const {
    return "NetworkLinkCorruptor";
}

void NetworkLinkCorruptor::outputClassHierarchy(ostream &out) const {
    this->NetworkLink::outputClassHierarchy(out);
    out << " -> " << this->NetworkLinkCorruptor::className();
}

void NetworkLinkCorruptor::getDescription(ostream &out) const {
    this->NetworkLink::getDescription(out);
    out << "Bit toggle probability: " << bitToggleProbability << '\n';
}

void NetworkLinkCorruptor::encode(Coder *coder) const {
    this->NetworkLink::encode(coder);
    
    coder->encodeFloat(bitToggleProbability);
}

bool NetworkLinkCorruptor::decode(Decoder *decoder) {
    if (this->NetworkLink::decode(decoder)) {
        bitToggleProbability = decoder->decodeFloat();
                
        return true;
    } else {
        return false;
    }
}
