//  NetworkLinkCorruptor.h
//  Network Simulator
//
//  Created by Wade Tregaskis on Thu Apr 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NETWORK_LINK_CORRUPTOR_H
#define NETWORK_LINK_CORRUPTOR_H

#include "NetworkLink.h"


#define NETWORK_LINK_CORRUPTOR 30


#define RANDOM(min, max) ((min) + (((max) - (min)) * ((float)random() / RAND_MAX)))


/*! @class NetworkLinkCorruptor
    @abstract Represents a network link which randomly - but probabilistically - corrupts data.
    @discussion The NetworkLinkCorruptor object receives data from one or more sources and forwards it, after possibly corrupting it to some extent, to the outlet.  Note that a NetworkLinkCorruptor, as a subclass of NetworkLink, may have one and only one outlet.  Note also that, again as a subclass of NetworkLink, the output is written to the outlet regardless of whether or not the outlet is willing to accept it, so data may be lost if the outlet does not handle it.
    */

class NetworkLinkCorruptor : public NetworkLink {
protected:
    float bitToggleProbability; // The probability that each bit passing through the corruptor will be toggled.
    
/*! @function toggleBits
    @abstract Randomly toggles bits in the given data.
    @discussion This member function is used when data is first submitted to toggle it as necessary to fulfil the corruptor's primary function.
    @param discussion data The data to [potentially] corrupt.
    @param length The length of the data. */
    
    void toggleBits(char *data, unsigned int length);
    
public:
/*! @function NetworkLinkCorruptor
    @abstract Designated initialiser for the NetworkLinkCorruptor class.
    @discussion Initialises the receiver in the given network with the given properties.
    @param transmission The transmission time for data across the link.  A transmission time of 0 has the link pass all data as soon as it is received (i.e. in the same tick).  A transmission time of 1 has it pass data out on the tick after it receives it, etc etc.  Must be greater than or equal to 0.
    @param bitToggleProb The probability of an individual bit being toggled (as a simulation of data corruption during transmission).  Must be greater than or equal to 0.0.  0.0 represents no chance, while 1.0 represents certainty - every single bit will be toggled. */

    NetworkLinkCorruptor(Network *home = NULL, int transmission = 0, float bitToggleProb = 0.0);
    ~NetworkLinkCorruptor(); // Standard black box destructor... nothing to see here.
    
    // NetworkEntity
        
    void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership);
    
    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};

#endif
