//  NetworkReceiverStopWait.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Sun Apr 04 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "NetworkReceiverStopWait.h"

void NetworkReceiverStopWait::submitToAllReplets(const char *data, unsigned int length, bool &ownership, NetworkEntity *exception) {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    for (iter = replets.begin(), stopper = replets.end(); iter != stopper; ++iter) {
        if (*iter != exception) {
            (*iter)->submitData(data, length, this, ownership);
        }
    }
}

NetworkReceiverStopWait::NetworkReceiverStopWait(Network *home) : NetworkEntity(home), sequence(0) {
    this->addType(NETWORK_RECEIVER_STOP_WAIT);
}

NetworkReceiverStopWait::~NetworkReceiverStopWait() {
    vector<NetworkEntity*>::iterator iter, stopper;

    for (iter = replets.begin(), stopper = replets.end(); iter != stopper; ++iter) {
        RELEASE(*iter);
    }
}

bool NetworkReceiverStopWait::addReplet(NetworkEntity *replet) {
    vector<NetworkEntity*>::const_iterator iter, stopper;
    
    if (replet && (replet != this)) {
        for (iter = replets.begin(), stopper = replets.end(); iter != stopper; ++iter) {
            if (replet == *iter) {
                return true;
            }
        }
        
        RETAIN(replet);
        replets.push_back(replet);
        replet->setNetwork(network);
        
        return true;
    } else {
        return false;
    }
}

void NetworkReceiverStopWait::removeReplet(NetworkEntity *replet) {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    for (iter = replets.begin(), stopper = replets.end(); iter != stopper; ++iter) {
        if (*iter == replet) {
            replets.erase(iter);
            RELEASE(replet);
            return;
        }
    }
}

// NetworkEntity

bool NetworkReceiverStopWait::canAcceptData(int length, NetworkEntity *source) {
    return true;
}

void NetworkReceiverStopWait::submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) {
    // 1 [7] Type (1 == data, 0 == ack) : [6] Sequence flag : {([5] Response (1 = ok, 0 = resend) : [4..0] Reserved) | [5..0] Length }
    // 2-3 CRC16
    // 4+ Data
    
    char response;
    bool responseOwnership = false;

    if (length < 4) {
        printf("(%s) NetworkReceiverStopWait[%llu] received data of length %d, too short to be a meaningful data packet.  Sending NACK.\n", this->className(), id, length);
        
        response = sequence << 6;
    } else {
        bool type = data[0] & 0x80;
        int seq = (data[0] & 0x40) >> 6;

        if (type) {
            unsigned int dataLength = (data[0] & 0x3F) + 1;
            UInt16 crc = ((((UInt16)data[1]) << 8) & 0xFF00) | (((UInt16)data[2]) & 0xFF);

            if ((dataLength + 3) > length) {                
                printf("(%s) NetworkReceiverStopWait[%llu] did not receive enough data (%d bytes) - requires %d bytes for paylod of %d bytes.  Sending NACK for sequence #%d.\n", this->className(), id, length, (dataLength + 3), dataLength, (int)(data[0] & 0x40) >> 6);
                
                response = data[0] & 0x40;
            } else {                
                if ((dataLength + 3) < length) {
                    printf("(%s) NetworkReceiverStopWait[%llu] given more data than necessary in submitData.  Ignoring extra data.\n", this->className(), id);
                }
                
                if (seq != sequence) {                    
                    printf("(%s) NetworkReceiverStopWait[%llu] received data of sequence #%d when it was expecting #%d.  Sending NACK for #%d.\n", this->className(), id, seq, sequence, sequence);
                    
                    response = sequence << 6;
                } else {
                    UInt16 checkCRC = crc16OfString(data + 3, dataLength), backUp;
                    
                    checkCRC = crc16(checkCRC, (((UInt16)data[0]) << 8) & 0xFF00);
                    backUp = crc16(checkCRC, 0);
                    
                    //printf("Orig CRC = %hu, calc'd CRC = %hu, result = %hu.\n", crc, checkCRC, crc16(checkCRC, crc));
                    
                    checkCRC = crc16(checkCRC, crc);
                    
                    if (checkCRC != 0) {                        
                        printf("(%s) NetworkReceiverStopWait[%llu] received data with checksum of %u who's actual checksum is %u.  Sending NACK.\n", this->className(), id, crc, backUp);

                        response = data[0] & 0x40;
                    } else {
                        printf("(%s) NetworkReceiverStopWait[%llu] received valid data {seq = #%d, crc = %u, length = %d}.  Sending ACK.\n", this->className(), id, (data[0] & 0x40) >> 6, crc, dataLength);
                        
                        this->submitToAllOutlets(data + 3, dataLength, ownership);
                        
                        sequence ^= 1;
                        
                        response = (data[0] & 0x40) ^ 0x40;
                        response |= 0x20;
                    }
                }
            }
        } else {
            printf("(%s) NetworkReceiverStopWait[%llu] received response packet, rather than data packet.  Presuming corrupt data packet - sending NACK for sequence #%d.\n", this->className(), id, sequence);
            
            response = sequence << 1;
        }
    }
    
    this->submitToAllReplets(&response, 1, responseOwnership);
}

void NetworkReceiverStopWait::tick(void) {
    // Do nothing for the moment, although in future we might want to delay responses to simulate processing time
}

// Object

const char* NetworkReceiverStopWait::className(void) const {
    return "NetworkReceiverStopWait";
}

void NetworkReceiverStopWait::outputClassHierarchy(ostream &out) const {
    this->NetworkEntity::outputClassHierarchy(out);
    out << " -> " << this->NetworkReceiverStopWait::className();
}

void NetworkReceiverStopWait::getDescription(ostream &out) const {
    this->NetworkEntity::getDescription(out);
    
}

void NetworkReceiverStopWait::encode(Coder *coder) const {
    vector<NetworkEntity*>::const_iterator iter, stopper;
    
    this->NetworkEntity::encode(coder);
    
    coder->encodeInt(sequence);
    coder->encodeInt(replets.size());
    
    for (iter = replets.begin(), stopper = replets.end(); iter != stopper; ++iter) {
        coder->encodeObject(*iter);
    }
}

bool NetworkReceiverStopWait::decode(Decoder *decoder) {
    if (this->NetworkEntity::decode(decoder)) {
        int count;
        NetworkEntity *temp;
        
        sequence = decoder->decodeInt();
        
        count = decoder->decodeInt();
        
        if (count < 0) {
            fprintf(stderr, "Error: (%s) NetworkReceiverStopWait[%llu] decoding failed - invalid number of replets \"%d\".\n", this->className(), id, count);
            return false;
        }
        
        while (count--) {
            temp = dynamic_cast<NetworkEntity*>(decoder->decodeObject());
            
            if (!temp) {
                fprintf(stderr, "Error: (%s) NetworkReceiverStopWait[%llu] decoding failed - unable to decode replet number %d.\n", this->className(), id, replets.size());
                return false;
            }
            
            replets.push_back(temp);
        }
        
        return true;
    } else {
        return false;
    }
}
