//  NetworkReceiverStopWait.h
//  Network Simulator
//
//  Created by Wade Tregaskis on Sun Apr 04 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NETWORK_RECEIVER_STOP_WAIT_H
#define NETWORK_RECEIVER_STOP_WAIT_H

#include "Network.h"
#include "CRC.h"


#define NETWORK_RECEIVER_STOP_WAIT 11


/*! @class NetworkReceiverStopWait
    @abstract Represents a data receiver which uses the StopWait protocol.
    @discussion Complementary [obviously] to the NetworkTransmitterStopWait class, this receiving end expects data to be sent to it, which it then verifies as necessary and then passes on to it's outlets, as required.  Unlike basic NetworkEntity's it requires a reply link to send back acknowledgements or errors, which is connected using the addReplet member function.  See the appropriate documentation for more details.
                As an overview, the StopWait protocol uses a very simple mechanism for transmitting data in a reliable fashion.  Only one packet may be in flight between the transmitter and receiver at any time - the receiver must acknowledge that packet before another can be sent.  Nonetheless each packet has a sequence number, either 0 or 1, which while not strictly necessary for the StopWait protocol does add some degree of reliability to the protocol in some extreme circumstances.
                Each packet also includes a CRC which the receiver uses to verify the integrity of the packet and it's payload.  If the CRC verification fails, the packet is rejected and an appropriate error response sent back to the transmitter.  The transmitter keeps resending the same packet until the transmission is successful, at which point it can move on to the next packet to be sent.
                The transmitter also includes a timeout mechanism so that if a packet is sent but no reply receives within a given time limit, the packet is assumed to have been lost and the transmitter resends it.

                To use the receiver, connect it as an outlet to whatever NetworkEntity carriers incoming data.  Add as a replet one or more links which end up back at the transmitter.  Add as outlets any NetworkEntity's you want to receive the received data (having been verified as uncorrupted).
    */

class NetworkReceiverStopWait : public NetworkEntity {
protected:
    int sequence; // The current sequence number, either 0 or 1
    vector<NetworkEntity*> replets; // The reply outlet(s) to send packet responses to
 
/*! @function submitToAllReplets
    @abstract Submits the given data to all the receiver's replets.
    @discussion Analogous to the submitToAllOutlets member function (of the NetworkEntity class), this version sends to the receiver's replets instead.  See the documentation for the submitToAllOutlets member function for details.
    @param data The data to transmit.
    @param length The length of said data.
    @param ownership On input: true if the given data may be claimed by one of the replets, in which case that claiming replet becomes responsible for freeing it as appropriate.  If false the data is owned by the caller and replets must copy it if they need prolonged access to it.  On output [only meaningful if true on input]: true if the caller still owns the given data, false if one of the replets claimed it.  The caller is still responsible for freeing the data in the former case.
    @param exception The NetworkEntity to <i>not</i> submit the given data to, analagous to the same parameter of the submitToAllOutlets function, albeit usually less useful. */
    
    virtual void submitToAllReplets(const char *data, unsigned int length, bool &ownership, NetworkEntity *exception = NULL);

public:
/*! @function NetworkReceiverStopWait
    @abstract Designated initialiser for the NetworkReceiverStopWait class.
    @discussion Initialises the receiver as part of the given network, if any.
    @param home The network to automatically add the receiver to, if any. */
        
    NetworkReceiverStopWait(Network *home = NULL);
    ~NetworkReceiverStopWait(); // Standard block box destructor... nothing to see here.
    
/*! @function addReplet
    @abstract Adds a given NetworkEntity as a replet (reply outlet) of the receiver.
    @discussion See the class discussion for details on the purpose of replets, and the discussion for NetworkEntity's addOutlet function for analogous details.
    @param replet The replet to add to the receiver.
    @result Returns true if the given NetworkEntity was successfully added as a replet to the receiver [or already was], false otherwise. */
    
    virtual bool addReplet(NetworkEntity *replet);
    
/*! @function removeReplet
    @abstract Removes a given NetworkEntity from the receiver's list of replets (reply outlets).
    @discussion See the class discussion for details on the purpose of replets, and the discussion for NetworkEntity's addOutlet function for analogous details.
    @param replet The replet to remove from the receiver. */
    
    virtual void removeReplet(NetworkEntity *replet);
    
    // NetworkEntity
    
    bool canAcceptData(int length, NetworkEntity *source);
    void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership);
    void tick(void);
    
    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};

#endif
