//  NetworkTransmitterStopWait.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Sat Apr 03 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "NetworkTransmitterStopWait.h"


NetworkTransmitterStopWait::NetworkTransmitterStopWait(Network *home, int timeout) : NetworkEntity(home), sequence(0), transmissionTimeout(timeout), currentTimer(0), ready(true), packet(NULL) {
    this->addType(NETWORK_TRANSMITTER_STOP_WAIT);
}

NetworkTransmitterStopWait::~NetworkTransmitterStopWait() {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    for (iter = inlets.begin(), stopper = inlets.end(); iter != stopper; ++iter) {
        RELEASE(*iter);
    }
    
    if (packet) {
        free(packet);
    }
}

bool NetworkTransmitterStopWait::addInlet(NetworkEntity *inlet) {
    vector<NetworkEntity*>::iterator iter, stopper;

    for (iter = inlets.begin(), stopper = inlets.end(); iter != stopper; ++iter) {
        if (*iter == inlet) {
            return true;
        }
    }
    
    RETAIN(inlet);
    inlets.push_back(inlet);
    
    return true;
}

void NetworkTransmitterStopWait::removeInlet(NetworkEntity *inlet) {
    vector<NetworkEntity*>::iterator iter, stopper;
    NetworkEntity *temp;
    
    for (iter = inlets.begin(), stopper = inlets.end(); iter != stopper; ++iter) {
        temp = *iter;
        
        inlets.erase(iter);
        
        RELEASE(temp);
        
        return;
    }
}

bool NetworkTransmitterStopWait::finished(void) {
    return (packet == NULL);
}


// NetworkEntity

bool NetworkTransmitterStopWait::canAcceptData(int length, NetworkEntity *source) {
    vector<NetworkEntity*>::iterator iter, stopper;
    
    if (!ready || (length > 64) || (length <= 0)) {
        return false;
    }
    
    for (iter = outlets.begin(), stopper = outlets.end(); iter != stopper; ++iter) {
        if (!(*iter)->canAcceptData(length + 3, source)) {
            return false;
        }
    }
    
    return true;
}

void NetworkTransmitterStopWait::submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership) {
    // 1 [7] Type (1 == data, 0 == ack) : [6] Sequence flag : {([5] Response (1 = ok, 0 = resend) : [4..0] Reserved) | [5..0] Length }
    // 2-3 CRC16
    // 4+ Data

    if (vectorContains(&inlets, source)) {
        if (length < 1) {
            printf("(%s) NetworkTransmitterStopWait[%llu] received incomplete data, length of %d.  Ignoring.\n", this->className(), id, length);
        } else {
            bool type = data[0] & 0x80;
            int seq = (data[0] & 0x40) >> 6;
            
            if (type) {
                printf("(%s) NetworkTransmitterStopWait[%llu] received supposed data packet - that's the job of a receiver.\n", this->className(), id);
            } else {
                bool ok = data[0] & 0x20;
                
                if (length > 1) {
                    printf("(%s) NetworkTransmitterStopWait[%llu] receives more than data of length %d from inlet, which is more than the single byte necessary for an ACK/NACK.  Ignoring excess data.\n", this->className(), id, length);
                }

                if (ok) {
                    if (seq == sequence) {
                        printf("(%s) NetworkTransmitterStopWait[%llu] received ACK indicating ready for packet #%d.\n", this->className(), id, seq);
                        ready = true;
                        
                        if (packet) {
                            free(packet);
                            packet = NULL;
                        }
                        
                        currentTimer = 0;
                    } else {
                        printf("(%s) NetworkTransmitterStopWait[%llu] received ACK indicating ready for packet #%d, but the last packet sent was #%d.  Ignoring.\n", this->className(), id, seq, sequence ^ 1);
                    }
                } else { // need to resend
                    bool ownership = false;
                    
                    if (seq == sequence) {
                        printf("(%s) NetworkTransmitterStopWait[%llu] received NACK for packet #%d, but the last packet sent was #%d.  Ignoring.\n", this->className(), id, seq, seq ^ 1);
                    } else {
                        if (packet) {
                            printf("(%s) NetworkTransmitterStopWait[%llu] received NACK for packet #%d.  Resending.\n", this->className(), id, seq);
                            
                            currentTimer = transmissionTimeout;
                            
                            this->submitToAllOutlets(packet, lastLength, ownership);
                        } else {
                            printf("(%s) NetworkTransmitterStopWait[%llu] received NACK for packet #%d, but this has already been ACK'd [or was never sent in the first place].  Ignoring.\n", this->className(), id, seq);
                        }
                    }
                }
            }
        }
    } else {
        if (!ready) {
            fprintf(stderr, "(%s) NetworkTransmitterStopWait[%llu] given data via submitData even though it is not ready.  Ignoring.\n", this->className(), id);
        } else if ((length <= 0) || (length > 64)) {
            fprintf(stderr, "(%s) NetworkTransmitterStopWait[%llu] given data greater than 64 bytes in length (or 0) - %d.  Ignoring.\n", this->className(), id, length);
        } else {
            UInt16 crc;
            bool myOwnership = false;

            if (packet) {
                free(packet);
            }
            
            lastLength = length + 3;
            
            packet = (char*)malloc(lastLength);

            packet[0] = (char)(length - 1);
            packet[0] |= 0x80;
            packet[0] |= ((sequence << 6) & 0x40); // Just to make sure if sequence is outside the range [0..1] it doesn't overwrite anything... hopefully it won't be able to get outside that range anyway

            memcpy(packet + 3, data, length);
            
            crc = crc16OfString(packet + 3, length);
            crc = crc16(crc, (((UInt16)packet[0]) << 8) & 0xFF00); // Update the CRC to include the header
            crc = crc16(crc, 0);
            
            packet[1] = (unsigned char)((crc & 0xFF00) >> 8);
            packet[2] = (unsigned char)(crc & 0xFF);

            printf("(%s) NetworkTransmitterStopWait[%llu] transmitting data {seq = #%d, crc = %u, length = %d}.\n", this->className(), id, sequence, crc, length);
            
            sequence ^= 1;
            ready = false;
            
            currentTimer = transmissionTimeout;
            
            this->submitToAllOutlets(packet, lastLength, myOwnership);
        }
    }
}

void NetworkTransmitterStopWait::tick(void) {
    --currentTimer;
    
    if (0 == currentTimer) {
        bool ownership = false;

        printf("(%s) NetworkTransmitterStopWait[%llu] timed out waiting for response to packet #%d.  Resending.\n", this->className(), id, (packet[0] & 0x40) >> 6);
        
        this->submitToAllOutlets(packet, lastLength, ownership);
        
        currentTimer = transmissionTimeout;
    }
}

// Object

const char* NetworkTransmitterStopWait::className(void) const {
    return "NetworkTransmitterStopWait";
}

void NetworkTransmitterStopWait::outputClassHierarchy(ostream &out) const {
    this->NetworkEntity::outputClassHierarchy(out);
    out << " -> " << this->NetworkTransmitterStopWait::className();
}

void NetworkTransmitterStopWait::getDescription(ostream &out) const {
    this->NetworkEntity::getDescription(out);
    
    out << "Next sequence #: " << sequence << "\nReady: " << (ready ? "yes" : "no");
    
    if (!inlets.empty()) {
        vector<NetworkEntity*>::const_iterator iter, stopper;
        
        out << "\nInlets: ";
        
        for (iter = inlets.begin(), stopper = inlets.end(); iter != stopper; ++iter) {
            out << (*iter)->getID();
            
            if ((iter + 1) != stopper) {
                out << ", ";
            }
        }
    }
    
    if (packet) {
        out << "\nLast packet: {seq = #" << ((packet[0] & 0x40) >> 6) << ", crc = " << (((UInt16)packet[1] << 8) + (UInt16)packet[2]) << ", length = " << lastLength << '}';
    }
    
    out << '\n';
}

void NetworkTransmitterStopWait::encode(Coder *coder) const {
    vector<NetworkEntity*>::const_iterator iter, stopper;
    
    this->NetworkEntity::encode(coder);
    
    coder->encodeInt(sequence);
    coder->encodeBool(ready);
    coder->encodeInt(transmissionTimeout);
    
    coder->encodeInt(inlets.size());
    
    for (iter = inlets.begin(), stopper = inlets.end(); iter != stopper; ++iter) {
        coder->encodeObject(*iter);
    }
    
    coder->encodeData(packet, (packet ? lastLength : 0));
}

bool NetworkTransmitterStopWait::decode(Decoder *decoder) {
    if (this->NetworkEntity::decode(decoder)) {
        int temp;
        NetworkEntity *current;
        
        sequence = decoder->decodeInt();
        ready = decoder->decodeBool();
        transmissionTimeout = decoder->decodeInt();
        
        temp = decoder->decodeInt();
        
        if (temp < 0) {
            fprintf(stderr, "Error: (%s) NetworkTransmitterStopWait[%llu] decoding failed - invalid number of outlets \"%d\".\n", this->className(), id, temp);
            return false;
        }
        
        while (temp--) {
            current = dynamic_cast<NetworkEntity*>(decoder->decodeObject());
            
            if (!current) {
                fprintf(stderr, "Error: (%s) NetworkTransmitterStopWait[%llu] decoding failed - unable to decode inlet #%d.\n", this->className(), id, inlets.size());
                return false;
            }
            
            inlets.push_back(current);
        }
        
        packet = decoder->decodeData(lastLength);
        
        return true;
    } else {
        return false;
    }
}
