//  NetworkTransmitterStopWait.h
//  Network Simulator
//
//  Created by Wade Tregaskis on Sat Apr 03 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef NETWORK_TRANSMITTER_STOP_WAIT_H
#define NETWORK_TRANSMITTER_STOP_WAIT_H

#include "CRC.h"
#include "Network.h"
#include "VectorAdditions.h"


#define NETWORK_TRANSMITTER_STOP_WAIT 10


/*! @class NetworkTransmitterStopWait
    @abstract Represents a data transmitter which uses the StopWait protocol.
    @discussion Complementary [obviously] to the NetworkReceiverStopWait class, this transmitting end sends data towards [hopefully] a receiver, which then replies with an appropriate response based on how successful the transmission was.
                As an overview, the StopWait protocol uses a very simple mechanism for transmitting data in a reliable fashion.  Only one packet may be in flight between the transmitter and receiver at any time - the receiver must acknowledge that packet before another can be sent.  Nonetheless each packet has a sequence number, either 0 or 1, which while not strictly necessary for the StopWait protocol does add some degree of reliability to the protocol in some extreme circumstances.
                Each packet also includes a CRC which the receiver uses to verify the integrity of the packet and it's payload.  If the CRC verification fails, the packet is rejected and an appropriate error response sent back to the transmitter.  The transmitter keeps resending the same packet until the transmission is successful, at which point it can move on to the next packet to be sent.
                The transmitter also includes a timeout mechanism so that if a packet is sent but no reply receives within a given time limit, the packet is assumed to have been lost and the transmitter resends it.

                To use the transmitter, connect it as an outlet to whatever NetworkEntity carriers incoming data, and inform the transmitter of this fact using it's addInlet member function.  Add as an outlet any link(s) which end up at the receiver.  Add the transmitter as an outlet of any links which carry response data back from the receiver.
*/

class NetworkTransmitterStopWait : public NetworkEntity {
private:
    int sequence, lastLength, transmissionTimeout, currentTimer; // Sequence records the 
    bool ready;
    vector<NetworkEntity*> inlets;
    char *packet;
    
public:
/*! @function NetworkTransmitterStopWait
    @abstract Designated initialiser for the NetworkTransmitterStopWait class.
    @discussion Initialises the receiver as a part of the given network (if any), with the given transmission timeout (if any).
    @param home The network to automatically add the receiver to, which may be NULL.
    @param timeout The number of cycles to wait after transmission of a frame before assuming it was lost and resending.  This must be larger than the minimum response time, otherwise you'll create strange problems and may not be able to transfer any data.  If it is 0 it is ignored - transmission never times out, so a single lost frame will halt the transmission indefinitely. */
        
    NetworkTransmitterStopWait(Network *home = NULL, int timeout = 0);
    ~NetworkTransmitterStopWait(); // Standard black box destructor... nothing to see here.
    
/*! @function addInlet
    @abstract Adds a given NetworkEntity as an inlet into the receiver.
    @discussion Inlets work in the opposite fashion to outlets - they are the paths through which data is expected to arrive to the receiver.  They are not normally specified explicitely because there is no need.  In the case of the transmitter, however, it needs to be able to tell the difference between data it should transmit and response coming back from the receiver.  The latter should come in via an inlet.
                Note that the given NetworkEntity will <i>not</i> have the receiver automatically added to it as an outlet.  This is to allow the network creator some flexibility - just because object A is an inlet of object B does not necessarily mean you always want to transmit data from A to B.
    @param inlet The NetworkEntity to mark as an inlet of the receiver.
    @result True if the given NetworkEntity was successfully added as an inlet of the receiver (or already was), false otherwise. */
                
    virtual bool addInlet(NetworkEntity *inlet);
    
/*! @function removeInlet
    @abstract Removes a given NetworkEntity from the receiver's list of inlets.
    @discussion See the discussion of the addInlet member function for more details.  Note that the given NetworkEntity will not automatically have the receiver removed from it's list of outlets - it is up to the network manager to decide whether this behaviour is desired or not, and if so to perform it explicitly.
    @param inlet The NetworkEntity to remove from the receiver's list of inlets. */
    
    virtual void removeInlet(NetworkEntity *inlet);

/*! @function finished
    @abstract Returns whether or not the receiver has finished sending all queued packets.
    @discussion This returns true if the receiver has no packets currently buffered.  This does not mean transmission is finished, just that it hasn't received any further data from it's inlets since the last segment of data was transmitted successfully.
    @result True if the receiver has finished transmitting all queued packets (and has at present no more data to send), false otherwise. */
    
    bool finished(void);
    
    // NetworkEntity
    
    bool canAcceptData(int length, NetworkEntity *source);
    void submitData(const char *data, unsigned int length, NetworkEntity *source, bool &ownership);
    void tick(void);
    
    // Object
    
    const char* className(void) const;
    void outputClassHierarchy(ostream &out) const;
    void getDescription(ostream &out) const;
    void encode(Coder *coder) const;
    bool decode(Decoder *decoder);
};

#endif

