//  Object.h
//
//  IMPORTANT: This code was developed by Wade Tregaskis under the employment of the Department of Primary Industries, Victoria, whom reserve all rights to it.  It has been used in good faith by Wade Tregaskis without explicit permission, but with the understanding from the employer that it's private use is acceptable in good faith.
//
//  The underlying message being, don't give this to anyone please as I don't have the rights for it's distribution.  Also, although I did write every character of it, I did so outside of university and as such I understand it may not be eligible for inclusion in university marking and other evaluation.
//
//  P.S. Half this code is redundant for this project, anyway.  I've included it in it's entirety only for completeness.

#ifndef OBJECT_H
#define OBJECT_H

#include <stdlib.h>

#include <map>
#include <set>

#include <iostream>
#include <assert.h>
#include <fstream>
using namespace std;


/*! @header Object
    @abstract Contains the Object, Coder and Decoder classes - the Spatial library foundation.
    */

class Object;


#define OBJECT 0
    

/*! @class Coder
    @abstract Represents a coding stream to some file.
    @discussion Coder's are created with particular files as their output.  They can then have their various member functions called to code data to that file, including Object's.  The corresponding Decoder class provides facilities for extracting the information back from the file and restoring the data/Object's in memory.
    */
    
class Coder {
protected:
    ofstream output;
    set<const Object*> encodedObjects; // We maintain a list of already encoded objects.  If we are asked to encode an object that is in this list, we can just write the ID of the object and not worry about encoding it again.  This of course has the very significant added bonus of allowing for circular references.

public:
/*! @function Coder
    @abstract Designated initialiser for the Coder class.
    @discussion This initialiser provides the coder with a destination for it's coded output.  If the file at the path given already exists it will be overwritten.  The file remains open for the life of the Coder - you should make sure you delete the Coder as soon as you are finished with it.
    @param path The path of the file to output the coded data to. */
    
    Coder(const char *path) : output(path, ios::binary | ios::out) {}

/*! @function ~Coder
    @abstract Destructor for the Coder class.
    @discussion This does nothing in the base class but act as a required placeholder. */
    
    virtual ~Coder() {}

/*! @function encodeBool
    @abstract Encodes a boolean value.
    @discussion The boolean value given is coded and written to the receiver's file.
    @param value The value to encode. */
    
    virtual void encodeBool(bool value);

/*! @function encodeInt
    @abstract Encodes an integer value.
    @discussion The integer value given is coded and written to the receiver's file.
    @param value The value to encode. */
    
    virtual void encodeInt(int value);

/*! @function encodeFloat
    @abstract Encodes a floating point value.
    @discussion The floating point value given is coded and written to the receiver's file.
    @param value The value to encode. */

    virtual void encodeFloat(float value);

/*! @function encodeString
    @abstract Encodes a string.
    @discussion The string given is coded and written to the receiver's file.  The string should be NULL-terminated.
    @param value The value to encode. */
    
    virtual void encodeString(const char *value);

/*! @function encodeData
    @abstract Encodes some amount of raw data.
    @discussion Similar to the encodeString class method, except that encodeData does not stop at the first NULL - it encodes the number of bytes instructed, regardless of the structure of the data.
    @param data The data to encode.  If length is greater than zero, this must not be NULL - if it is an assertion will fail and the app will crash.  If length is equal to 0, this may be NULL or otherwise; it will be ignored.
    @param length The length of the data to encode.  This must be equal to or greater than zero - if it is less than an assertion will fail and the app will crash. */
    
    virtual void encodeData(const char *data, int length);
    
/*! @function encodeObject
    @abstract Encodes an Object.
    @discussion This is the real godsend of the Coder/Decoder system.  The given object is written to the file by encoding it's address and ID, followed by it's class specific information as encoded using the Object class's member function encode().  If the given object contains other objects [or references to], it may well invoke this member function recursively.
    
                       Note that each object is encoded just once.  If asked to encode an object a second time, the receiver records only a reference to the already-coded version earlier in the file.  In this manner infinite loops are avoided, not to mention the overhead of encoding the same object multiple times.
    @param value The Object to encode. */
    
    virtual void encodeObject(const Object *value);
};


/*! @class Decoder
    @abstract Represents a decoding stream from some file.
    @discussion Decoder's are created with particular files as their input.  They can then have their various member functions called to decode data from that file, including Object's.  The corresponding Coder class provides facilities for writing the information to the file in the first place.
    */
    
class Decoder {
protected:
    ifstream input;
    map<void*, Object*> decodedObjects; // Every object decoded gets placed into this map, so that when we find a reference to an object we've already decoded we can just returned the existing instance, rather than creating a new one.  This saves both file space and serious headaches with value-identical but pointer-differentiate instances.  It also allows for circular references.

    void* (*objectMaker)(unsigned int);

public:
/*! @function Decoder
    @abstract Designated initialiser for the Decoder class.
    @discussion This initialiser provides the decoder with a source for it's coded input.  If the file at the path does not exist all outputs from the decode methods will be NULL.  The file remains open for the life of the Decoder - you should make sure you delete the Decoder as soon as you are finished with it.
    @param path The path of the file to input the coded data from.
    @param maker A function for making Object's.  Since the Spatial library's Decoder cannot possibly know what classes people may be trying to instantiate, it leaves the initialisation to a helper method.  This helper method takes as it's sole argument a class type, and should then return a new instance of that class.  The Decoder then calls the decode() member function of the object to properly initialise it with values from the file.  If you do not have your own maker function, you can use the generic built-in one, called SpatialMakerFoundation(), defined in GenericMaker.h. */
    
    Decoder(const char *path, void* (*maker)(unsigned int)) : input(path, ios::binary | ios::in), objectMaker(maker) { assert(objectMaker); }

/*! @function ~Decoder
    @abstract Destructor for the Decoder class.
    @discussion This does nothing in the base class but act as a required placeholder. */
    
    virtual ~Decoder() {}

/*! @function decodeBool
    @abstract Decodes a boolean value.
    @discussion A boolean value is read from the file and returned.
    @result Returns the boolean value, or false if an error occurs. */
    
    virtual bool decodeBool(void);

/*! @function decodeInt
    @abstract Decodes an integer value.
    @discussion An integer value is read from the file and returned.
    @result Returns the integer value, or 0 if an error occurs. */
    
    virtual int decodeInt(void);

/*! @function decodeFloat
    @abstract Decodes a floating point value.
    @discussion A floating point value is read from the file and returned.
    @result Returns the floating point value, or 0.0 if an error occurs. */
    
    virtual float decodeFloat(void);

/*! @function decodeString
    @abstract Decodes a string.
    @discussion A string is read from the file and returned.
    @result Returns the string (which may be NULL), or NULL if an error occurs.  You are responsible for freeing the returned string when you are finished using it. */
    
    virtual char* decodeString(void);

/*! @function decodeData
    @abstract Decodes some raw data.
    @discussion The raw data previously encoded at this point is recovered and returned, with it's length stored in the out parameter 'length'.
    @param length After calling the length of the data is set in this parameter.  If an error occurs, the given value will be negative and NULL returned.
    @result Returns the raw data (which may be NULL), or NULL if an error occurs.  Use the length returned to determine if an error really occured or if the data actually should be NULL.  You are responsible for freeing the returned data when you are finished using it. */
    
    virtual char* decodeData(int &length);
        
/*! @function decodeObject
    @abstract Decodes an Object.
    @discussion The receiver extracts the stored Object's member class type and calls the maker function to initialise an instance of the class.  It then calls the new instance's decode() member function to initialise it from the file.  This process may be recursive - decoding an object may involve decoding one or more other objects, etc.  Note that since each object is stored only once in the file, requests for the same object multiple times will return the existing instance, not a new one.
    @result The decoded Object, or NULL if an error occurs. */
    
    virtual Object* decodeObject(void);
};

   
// These macro's aid both debugging (if enabled) and also in retain/release use.  Since they automagically handle the case where the given pointer is NULL, they remove a lot of the pain that might otherwise be caused by this.

#ifdef MDEBUG
#define RETAIN(x) cout << "Retain <" << (int)(x) << ">\n"; if ((x) != NULL) { (x)->retain(); }
#define RELEASE(x) cout << "Release <" << (int)(x) << ">\n"; if ((x) != NULL) { (x)->release(); }
#else
#define RETAIN(x) if ((x) != NULL) { (x)->retain(); }
#define RELEASE(x) if ((x) != NULL) { (x)->release(); }
#endif


/*! @class Object
     @abstract The foundation object.
     @discussion Provides the most fundamental building block for OOP design - a reference-counting dynamically-typed object.  Those who've used ObjC will already understand how fantastic this is.  This is of course nowhere near as good as the standard Objective-C implementation, due to the relative limitations of the C/C++ runtime, but it's a reasonable approximation of the fundamental features.
     */

class Object {
protected:
    unsigned int retainCount, *objectTypes, numberOfTypes, typeCapacity;


/*! @function addType
    @abstract Adds the given type to the receiver's list of types.
    @discussion Since a class can inherit from another, of course, each instance must maintain a list of types.  These form a hierarchy starting from the root Object type and descending through successive subtypes down to the 'true' type of the object.
    @param type The class type to add. */
    
    inline void addType(unsigned int type) {
        if (numberOfTypes == typeCapacity) {
            typeCapacity += 4;
            objectTypes = (unsigned int*)realloc(objectTypes, sizeof(unsigned int) * typeCapacity);
        }

        objectTypes[numberOfTypes++] = type;
    }

public:
/*! @function Object
     @abstract Designated initialiser for the Object class.
     @discussion Initialises the retain count to 1, i.e. the receiver is implicitly retained by the constructing object.  Also constructs the class type list and enters it's own class as the root entry. */

    Object();

/*! @function ~Object
     @abstract Standard virtual destructor.
     @discussion Serves no great purpose other than to ensure the destructors of subclasses are called properly, and also useful in debugging to see when objects are really destroyed. */

    virtual ~Object() { delete objectTypes; }

/*! @function isKindOfClass
    @abstract Returns whether or not the receiver is a kind of the given class.
    @discussion An object is a kindOf a class if it is an instance of that class or any of it's subclasses.
    @param type The class type.
    @result True if the receiver is a kindOf the given class, false otherwise. */
    
    inline bool isKindOfClass(unsigned int type) const {
        for (unsigned int i = 0; i < numberOfTypes; ++i) {
            if (type == objectTypes[i]) {
                return true;
            }
        }

        return false;
    }

/*! @function isMemberOfClass
    @abstract Returns whether or not the receiver is a member of the given class.
    @discussion An object is a memberOf a class if it is an instance of that exact class.  Subclasses of that class do not count (see isKindOfClass if that's what you're looking for).
    @param type The class type.
    @result True if the receiver is a memberOf the given class, false otherwise. */
    
    inline bool isMemberOfClass(unsigned int type) const {
        return (objectTypes[numberOfTypes - 1] == type);
    }

/*! @function memberClass
    @abstract Returns the deepest class type of the receiver.
    @discussion The class returned is that at the bottom of the class hierarchy, e.g. for an object with the hierarchy Object -> Object2D -> MutableObject2D -> Warren, the class type is Warren.  Note that you should not use the returned value for comparing classes, as of course the class type of different classes, even if they inherit from the same parent, will be different.  Use the isKindOfClass or isMemberOfClass member functions instead.
    @result Returns the deepest class type of the receiver. */
    
    inline unsigned int memberClass(void) const {
        return objectTypes[numberOfTypes - 1];
    }
    
/*! @function className
    @abstract Returns the name of the receiver's class.
    @discussion In normal unqualified use (i.e. without specifying which class's implementation to call explicitly) this returns the name of the lowest subclass of the receiver.
    @result The name of the receiver. */
        
    virtual const char* className(void) const {
        return "Object";
    }

/*! @function outputClassHierarchy
    @abstract Outputs the class hierarchy of the receiver.
    @discussion Each class type is output by name, separated by ' -> ', and listed in appropriate order from root to lowest subclass.
    @param out The output stream to write the hierarchy to. */
    
    virtual void outputClassHierarchy(ostream &out) const {
        out << this->Object::className();
    }

/*! @function retain
     @abstract Increases the receivers retain count by 1.
     @discussion Every time an object maintains a reference to another object, it should retain that object.  When the reference is destroyed, it should release the object.  An Object is only destroyed when it has been released as many times as it was retained, thus ensuring the object lives for as long as necessary, but no more. */

    inline void retain(void) { ++retainCount; }

/*! @function release
     @abstract Decrements the receivers retain count by 1, potentially deallocating it.
     @discussion If the retain count hits 0, the receiver is deallocated.  If release is called too many times, an exception is raised. */

    inline void release(void) {
        assert(retainCount > 0);
        
        --retainCount;
        
        if (0 == retainCount) {
            //cout << "Deleting " << (int)this << endl;
            
            delete this;
        }
    }

/*! @function getDescription
    @abstract Returns a description of the receiver.
    @discussion This member function should be overriden by each subclass, and implemented in such a way that the subclass calls it's parents implemention, then adds on any of it's own information.  This allows a hierarchal description of the object to be constructed, where each specific class need only add whatever information it specifically provides.
    @param out The output stream to write the description to. */
    
    virtual void getDescription(ostream &out) const;

/*! @function encode
    @abstract Encodes the receiver - that is, places all it's properties into the given dictionary.
    @discussion The purpose of this method is to allow Object's to be encoded as a set of keyed string values, which can then be written to file or otherwise manipulated.  The corresponding decode() member function takes the values back from a given dictionary and initialises the receiver appropriately.
    @param dictionary The dictionary (hash_map) into which the keyed values should be placed. */
    
    virtual void encode(Coder *coder) const;

/*! @function decode
    @abstract Decodes the given dictionary into the receiver.
    @discussion The receiver takes from the given dictionary all the values it understands and initialises itself appropriately.  If there are missing values the decoding may fail, depending on how important the missing values are.  In the case that important values are missing, or otherwise invalid, this method returns false and the receiver should be considered unusable.
    @param dictionary The dictionary to take values from.
    @result Returns true if the decode was successful, false otherwise.  If false the receiver should be considered unusable. */
    
    virtual bool decode(Decoder *decoder);
};

#endif
