//  main.cpp
//  Network Simulator
//
//  Created by Wade Tregaskis on Thu Apr 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "Network.h"
#include "NetworkFileSink.h"
#include "NetworkFileSource.h"
#include "NetworkLink.h"
#include "NetworkLinkCorruptor.h"
#include "NetworkLinkDropper.h"
#include "NetworkTransmitterStopWait.h"
#include "NetworkReceiverStopWait.h"

#include <iostream>
using namespace std;


// I've chosen to hard-wire the network configuration [in main()], despite the inherent flexibility of my system, simply because I don't have time to develop any proper interface for this software.  It obviously suits our particular use to a tee, and if I ever do want to reuse it, I can think about a configurable front-end then.  Or just rewrite the hard-coded stuff.  :)

int main(int argc, char *argv[]) {
    if (argc != 3) {
        fprintf(stderr, "Usage: NetworkSimulator INPUT OUTPUT\n\n\tINPUT\tThe input file to transmit on the simulated network.\n\tOUTPUT\tThe output file to write the transmitted data to.\n");
        exit(1);
    } else {
        Network *sampleNetwork;
        NetworkFileSource *source;
        NetworkFileSink *sink;
        NetworkLinkCorruptor *link;
        NetworkLinkDropper *link2;
        NetworkLink *linkBack;
        NetworkTransmitterStopWait *transmitter;
        NetworkReceiverStopWait *receiver;
        
        srandom(time(NULL));
        
        sampleNetwork = new Network();
        
        source = new NetworkFileSource(sampleNetwork, argv[1], 10); // Reads the file from disk, 10 bytes at a time
        sink = new NetworkFileSink(sampleNetwork, argv[2]); // Writes the file to disk
        
        transmitter = new NetworkTransmitterStopWait(sampleNetwork, 6); // Transmits each frame using the StopWait protocol, with a timeout of 6 ticks.
        receiver = new NetworkReceiverStopWait(sampleNetwork); // Receives each frame using the StopWait protocol
        
        link2 = new NetworkLinkDropper(sampleNetwork, 1, 0.01); // Want a 1% chance of dropping a frame completely.
        link = new NetworkLinkCorruptor(sampleNetwork, 1, 1 / 1040.0); // Assume every frame is 13 bytes long (full payload), = 104 bits, and want one error per 10 packets, therefore a one in 1040 chance of toggling a bit.
        linkBack = new NetworkLink(sampleNetwork, 1); // A latency of 1, although in this 'rigged' configuration this ends up being 0.  This vanilla link type is the one on which responses are returned to the transmitter - it is lossless as per requirements, since if it were to alter the data the StopWait protocol may not guarantee a sufficient level of reliability.
        
        source->addOutlet(transmitter); // Connect the file source to the transmitter
        transmitter->addOutlet(link2); // Connect the transmitter to the first link, the dropping link
        link2->addOutlet(link); // Connect the dropping link to the next link, the corrupting link
        link->addOutlet(receiver); // Connect the corrupting link to the receiver
        receiver->addOutlet(sink); // Connect to the receiver to the file sink
        
        receiver->addReplet(linkBack); // Add the reply link to the receiver as a replet (reply outlet)
        linkBack->addOutlet(transmitter); // Connect the reply link to the transmitter
        transmitter->addInlet(linkBack); // Configure the transmitter to see the reply link as such, rather than a data source (like the file source)
        
        printf("\nBeginning transmission of file \"%s\" over lossy, corrupting network, to file \"%s\".\n\n", argv[1], argv[2]);
        
        while (!(source->finished()) || !(transmitter->finished())) { // While there is more data to be sent from the file, or the transmitter still has a packet in flight, loop...
            sampleNetwork->simulate(1); // Simulate a single step of the simulation, i.e. one tick.
        }
        
        printf("\nTransmission of file complete.\n\n");
        
        // It's 'nice' to release memory explicitly, but there are two reasons why in this case there's no point:
        //  a) We know we're going to quit, so it's faster just to let the OS scrap our memory without worrying about the malloc details
        //  b) Because of the existince of some circular references in the various network components, releasing them all won't destroy them.  So, it's a bit of a pointless exercise.
        // The last point could be rectified by explicitly removing each outlet added (or at least, those necessary to destroy the circular references).  But, really, who cares, given (a).
        
        /*RELEASE(sampleNetwork);
        RELEASE(source);
        RELEASE(sink);
        RELEASE(link);
        RELEASE(link2);
        RELEASE(transmitter);
        RELEASE(receiver);
        RELEASE(linkBack);*/
    }
    
    return 0;
}
